@@ -1,14 +1,13 @@
 #include <Python.h>
 #include <time.h>
 
-#include "third-party/quickjs.h"
+#include "upstream-quickjs/quickjs.h"
 
 // Node of Python callable that the context needs to keep available.
 typedef struct PythonCallableNode PythonCallableNode;
 struct PythonCallableNode {
 	PyObject *obj;
-	// Internal ID of the callable function. "magic" is QuickJS terminology.
-	int magic;
+	PythonCallableNode *prev;
 	PythonCallableNode *next;
 };
 
@@ -27,24 +26,29 @@ typedef struct {
 	// Used when releasing the GIL.
 	PyThreadState *thread_state;
 	InterruptData interrupt_data;
-	// NULL-terminated singly linked list of callable Python objects that we need to keep alive.
+	// NULL-terminated doubly linked list of callable Python objects that we need to keep track of.
+	// We need to store references to callables in a place where we can access them when running
+	// Python's GC. Having them stored only in QuickJS' function opaques would create a dependency
+	// cycle across Python and QuickJS that neither GC can notice.
 	PythonCallableNode *python_callables;
-} ContextData;
+} RuntimeData;
 
 // The data of the type _quickjs.Object.
 typedef struct {
 	PyObject_HEAD;
-	ContextData *context;
+	RuntimeData *runtime_data;
 	JSValue object;
 } ObjectData;
 
 // The exception raised by this module.
 static PyObject *JSException = NULL;
 static PyObject *StackOverflow = NULL;
+// Converts the current Javascript exception to a Python exception via a C string.
+static void quickjs_exception_to_python(JSContext *context);
 // Converts a JSValue to a Python object.
 //
 // Takes ownership of the JSValue and will deallocate it (refcount reduced by 1).
-static PyObject *quickjs_to_python(ContextData *context_obj, JSValue value);
+static PyObject *quickjs_to_python(RuntimeData *runtime_data, JSValue value);
 
 // Returns nonzero if we should stop due to a time limit.
 static int js_interrupt_handler(JSRuntime *rt, void *opaque) {
@@ -57,54 +61,54 @@ static int js_interrupt_handler(JSRuntime *rt, void *opaque) {
 }
 
 // Sets up a context and an InterruptData struct if the context has a time limit.
-static void setup_time_limit(ContextData *context, InterruptData *interrupt_data) {
-	if (context->has_time_limit) {
-		JS_SetInterruptHandler(context->runtime, js_interrupt_handler, interrupt_data);
-		interrupt_data->limit = context->time_limit;
+static void setup_time_limit(RuntimeData *runtime_data, InterruptData *interrupt_data) {
+	if (runtime_data->has_time_limit) {
+		JS_SetInterruptHandler(runtime_data->runtime, js_interrupt_handler, interrupt_data);
+		interrupt_data->limit = runtime_data->time_limit;
 		interrupt_data->start = clock();
 	}
 }
 
 // Restores the context if the context has a time limit.
-static void teardown_time_limit(ContextData *context) {
-	if (context->has_time_limit) {
-		JS_SetInterruptHandler(context->runtime, NULL, NULL);
+static void teardown_time_limit(RuntimeData *runtime_data) {
+	if (runtime_data->has_time_limit) {
+		JS_SetInterruptHandler(runtime_data->runtime, NULL, NULL);
 	}
 }
 
 // This method is always called in a context before running JS code in QuickJS. It sets up time
 // limites, releases the GIL etc.
-static void prepare_call_js(ContextData *context) {
+static void prepare_call_js(RuntimeData *runtime_data) {
 	// We release the GIL in order to speed things up for certain use cases.
-	assert(!context->thread_state);
-	context->thread_state = PyEval_SaveThread();
-	setup_time_limit(context, &context->interrupt_data);
+	assert(!runtime_data->thread_state);
+	runtime_data->thread_state = PyEval_SaveThread();
+	setup_time_limit(runtime_data, &runtime_data->interrupt_data);
 }
 
 // This method is called right after returning from running JS code. Aquires the GIL etc.
-static void end_call_js(ContextData *context) {
-	teardown_time_limit(context);
-	assert(context->thread_state);
-	PyEval_RestoreThread(context->thread_state);
-	context->thread_state = NULL;
+static void end_call_js(RuntimeData *runtime_data) {
+	teardown_time_limit(runtime_data);
+	assert(runtime_data->thread_state);
+	PyEval_RestoreThread(runtime_data->thread_state);
+	runtime_data->thread_state = NULL;
 }
 
 // Called when Python is called again from inside QuickJS.
-static void prepare_call_python(ContextData *context) {
-	assert(context->thread_state);
-	PyEval_RestoreThread(context->thread_state);
-	context->thread_state = NULL;
+static void prepare_call_python(RuntimeData *runtime_data) {
+	assert(runtime_data->thread_state);
+	PyEval_RestoreThread(runtime_data->thread_state);
+	runtime_data->thread_state = NULL;
 }
 
 // Called when the operation started by prepare_call_python is done.
-static void end_call_python(ContextData *context) {
-	assert(!context->thread_state);
-	context->thread_state = PyEval_SaveThread();
+static void end_call_python(RuntimeData *runtime_data) {
+	assert(!runtime_data->thread_state);
+	runtime_data->thread_state = PyEval_SaveThread();
 }
 
 // GC traversal.
 static int object_traverse(ObjectData *self, visitproc visit, void *arg) {
-	Py_VISIT(self->context);
+	Py_VISIT(self->runtime_data);
 	return 0;
 }
 
@@ -112,19 +116,19 @@ static int object_traverse(ObjectData *self, visitproc visit, void *arg) {
 static PyObject *object_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
 	ObjectData *self = PyObject_GC_New(ObjectData, type);
 	if (self != NULL) {
-		self->context = NULL;
+		self->runtime_data = NULL;
 	}
 	return (PyObject *)self;
 }
 
 // Deallocates an instance of the Object class.
 static void object_dealloc(ObjectData *self) {
-	if (self->context) {
+	if (self->runtime_data) {
 		PyObject_GC_UnTrack(self);
-		JS_FreeValue(self->context->context, self->object);
-		// We incremented the refcount of the context when we created this object, so we should
+		JS_FreeValue(self->runtime_data->context, self->object);
+		// We incremented the refcount of the runtime data when we created this object, so we should
 		// decrease it now so we don't leak memory.
-		Py_DECREF(self->context);
+		Py_CLEAR(self->runtime_data);
 	}
 	PyObject_GC_Del(self);
 }
@@ -136,9 +140,9 @@ static PyObject *object_call(ObjectData *self, PyObject *args, PyObject *kwds);
 //
 // Returns the JSON representation of the object as a Python string.
 static PyObject *object_json(ObjectData *self) {
-	JSContext *context = self->context->context;
+	JSContext *context = self->runtime_data->context;
 	JSValue json_string = JS_JSONStringify(context, self->object, JS_UNDEFINED, JS_UNDEFINED);
-	return quickjs_to_python(self->context, json_string);
+	return quickjs_to_python(self->runtime_data, json_string);
 }
 
 // All methods of the _quickjs.Object class.
@@ -160,7 +164,7 @@ static PyTypeObject Object = {PyVarObject_HEAD_INIT(NULL, 0).tp_name = "_quickjs
                               .tp_methods = object_methods};
 
 // Whether converting item to QuickJS would be possible.
-static int python_to_quickjs_possible(ContextData *context, PyObject *item) {
+static int python_to_quickjs_possible(RuntimeData *runtime_data, PyObject *item) {
 	if (PyBool_Check(item)) {
 		return 1;
 	} else if (PyLong_Check(item)) {
@@ -173,7 +177,7 @@ static int python_to_quickjs_possible(ContextData *context, PyObject *item) {
 		return 1;
 	} else if (PyObject_IsInstance(item, (PyObject *)&Object)) {
 		ObjectData *object = (ObjectData *)item;
-		if (object->context != context) {
+		if (object->runtime_data != runtime_data) {
 			PyErr_Format(PyExc_ValueError, "Can not mix JS objects from different contexts.");
 			return 0;
 		}
@@ -190,7 +194,7 @@ static int python_to_quickjs_possible(ContextData *context, PyObject *item) {
 //
 // If the Python object is not possible to convert to JS, undefined will be returned. This fallback
 // will not be used if python_to_quickjs_possible returns 1.
-static JSValueConst python_to_quickjs(ContextData *context, PyObject *item) {
+static JSValueConst python_to_quickjs(RuntimeData *runtime_data, PyObject *item) {
 	if (PyBool_Check(item)) {
 		return JS_MKVAL(JS_TAG_BOOL, item == Py_True ? 1 : 0);
 	} else if (PyLong_Check(item)) {
@@ -200,18 +204,18 @@ static JSValueConst python_to_quickjs(ContextData *context, PyObject *item) {
 			PyObject *float_value = PyNumber_Float(item);
 			double double_value = PyFloat_AsDouble(float_value);
 			Py_DECREF(float_value);
-			return JS_NewFloat64(context->context, double_value);
+			return JS_NewFloat64(runtime_data->context, double_value);
 		} else {
 			return JS_MKVAL(JS_TAG_INT, value);
 		}
 	} else if (PyFloat_Check(item)) {
-		return JS_NewFloat64(context->context, PyFloat_AsDouble(item));
+		return JS_NewFloat64(runtime_data->context, PyFloat_AsDouble(item));
 	} else if (item == Py_None) {
 		return JS_NULL;
 	} else if (PyUnicode_Check(item)) {
-		return JS_NewString(context->context, PyUnicode_AsUTF8(item));
+		return JS_NewString(runtime_data->context, PyUnicode_AsUTF8(item));
 	} else if (PyObject_IsInstance(item, (PyObject *)&Object)) {
-		return JS_DupValue(context->context, ((ObjectData *)item)->object);
+		return JS_DupValue(runtime_data->context, ((ObjectData *)item)->object);
 	} else {
 		// Can not happen if python_to_quickjs_possible passes.
 		return JS_UNDEFINED;
@@ -220,7 +224,7 @@ static JSValueConst python_to_quickjs(ContextData *context, PyObject *item) {
 
 // _quickjs.Object.__call__
 static PyObject *object_call(ObjectData *self, PyObject *args, PyObject *kwds) {
-	if (self->context == NULL) {
+	if (self->runtime_data == NULL) {
 		// This object does not have a context and has not been created by this module.
 		Py_RETURN_NONE;
 	}
@@ -230,27 +234,36 @@ static PyObject *object_call(ObjectData *self, PyObject *args, PyObject *kwds) {
 	const int nargs = PyTuple_Size(args);
 	for (int i = 0; i < nargs; ++i) {
 		PyObject *item = PyTuple_GetItem(args, i);
-		if (!python_to_quickjs_possible(self->context, item)) {
+		if (!python_to_quickjs_possible(self->runtime_data, item)) {
 			return NULL;
 		}
 	}
 
 	// Now we know that all arguments are supported and we can convert them.
-	JSValueConst *jsargs = malloc(nargs * sizeof(JSValueConst));
+	JSValueConst *jsargs;
+	if (nargs) {
+		jsargs = js_malloc(self->runtime_data->context, nargs * sizeof(JSValueConst));
+		if (jsargs == NULL) {
+			quickjs_exception_to_python(self->runtime_data->context);
+			return NULL;
+		}
+	}
 	for (int i = 0; i < nargs; ++i) {
 		PyObject *item = PyTuple_GetItem(args, i);
-		jsargs[i] = python_to_quickjs(self->context, item);
+		jsargs[i] = python_to_quickjs(self->runtime_data, item);
 	}
 
-	prepare_call_js(self->context);
+	prepare_call_js(self->runtime_data);
 	JSValue value;
-	value = JS_Call(self->context->context, self->object, JS_NULL, nargs, jsargs);
+	value = JS_Call(self->runtime_data->context, self->object, JS_NULL, nargs, jsargs);
 	for (int i = 0; i < nargs; ++i) {
-		JS_FreeValue(self->context->context, jsargs[i]);
+		JS_FreeValue(self->runtime_data->context, jsargs[i]);
 	}
-	free(jsargs);
-	end_call_js(self->context);
-	return quickjs_to_python(self->context, value);
+	if (nargs) {
+		js_free(self->runtime_data->context, jsargs);
+	}
+	end_call_js(self->runtime_data);
+	return quickjs_to_python(self->runtime_data, value);
 }
 
 // Converts the current Javascript exception to a Python exception via a C string.
@@ -287,8 +300,8 @@ static void quickjs_exception_to_python(JSContext *context) {
 // Converts a JSValue to a Python object.
 //
 // Takes ownership of the JSValue and will deallocate it (refcount reduced by 1).
-static PyObject *quickjs_to_python(ContextData *context_obj, JSValue value) {
-	JSContext *context = context_obj->context;
+static PyObject *quickjs_to_python(RuntimeData *runtime_data, JSValue value) {
+	JSContext *context = runtime_data->context;
 	int tag = JS_VALUE_GET_TAG(value);
 	// A return value of NULL means an exception.
 	PyObject *return_value = NULL;
@@ -319,8 +332,8 @@ static PyObject *quickjs_to_python(ContextData *context_obj, JSValue value) {
 		ObjectData *object = (ObjectData *)return_value;
 		// This is important. Otherwise, the context may be deallocated before the object, which
 		// will result in a segfault with high probability.
-		Py_INCREF(context_obj);
-		object->context = context_obj;
+		Py_INCREF(runtime_data);
+		object->runtime_data = runtime_data;
 		PyObject_GC_Track(object);
 		object->object = JS_DupValue(context, value);
 	} else {
@@ -343,7 +356,7 @@ static PyObject *test(PyObject *self, PyObject *args) {
 struct module_state {};
 
 // GC traversal.
-static int context_traverse(ContextData *self, visitproc visit, void *arg) {
+static int runtime_traverse(RuntimeData *self, visitproc visit, void *arg) {
 	PythonCallableNode *node = self->python_callables;
 	while (node) {
 		Py_VISIT(node->obj);
@@ -354,7 +367,7 @@ static int context_traverse(ContextData *self, visitproc visit, void *arg) {
 
 // GC clearing. Object does not have a clearing method, therefore dependency cycles
 // between Context and Object will always be cleared starting here.
-static int context_clear(ContextData *self) {
+static int runtime_clear(RuntimeData *self) {
 	PythonCallableNode *node = self->python_callables;
 	while (node) {
 		Py_CLEAR(node->obj);
@@ -363,44 +376,129 @@ static int context_clear(ContextData *self) {
 	return 0;
 }
 
+static JSClassID js_python_function_class_id;
+
+static void js_python_function_finalizer(JSRuntime *rt, JSValue val) {
+	PythonCallableNode *node = JS_GetOpaque(val, js_python_function_class_id);
+	RuntimeData *runtime_data = JS_GetRuntimeOpaque(rt);
+	if (node) {
+		// fail safe
+		JS_SetOpaque(val, NULL);
+		// NOTE: This may be called from e.g. runtime_dealloc, but also from
+		// e.g. JS_Eval, so we need to ensure that we are in the correct state.
+		// TODO: integrate better with (prepare|end)_call_(python|js).
+		if (runtime_data->thread_state) {
+			PyEval_RestoreThread(runtime_data->thread_state);
+		}
+		if (node->prev) {
+			node->prev->next = node->next;
+		} else {
+			runtime_data->python_callables = node->next;
+		}
+		if (node->next) {
+			node->next->prev = node->prev;
+		}
+		// May have just been cleared in runtime_clear.
+		Py_XDECREF(node->obj);
+		PyMem_Free(node);
+		if (runtime_data->thread_state) {
+			runtime_data->thread_state = PyEval_SaveThread();
+		}
+	};
+}
+
+static JSValue js_python_function_call(JSContext *ctx, JSValueConst func_obj,
+                                       JSValueConst this_val, int argc, JSValueConst *argv,
+                                       int flags) {
+	RuntimeData *runtime_data = (RuntimeData *)JS_GetRuntimeOpaque(JS_GetRuntime(ctx));
+	PythonCallableNode *node = JS_GetOpaque(func_obj, js_python_function_class_id);
+	if (runtime_data->has_time_limit) {
+		return JS_ThrowInternalError(ctx, "Can not call into Python with a time limit set.");
+	}
+	prepare_call_python(runtime_data);
+
+	PyObject *args = PyTuple_New(argc);
+	if (!args) {
+		end_call_python(runtime_data);
+		return JS_ThrowOutOfMemory(ctx);
+	}
+	int tuple_success = 1;
+	for (int i = 0; i < argc; ++i) {
+		PyObject *arg = quickjs_to_python(runtime_data, JS_DupValue(ctx, argv[i]));
+		if (!arg) {
+			tuple_success = 0;
+			break;
+		}
+		PyTuple_SET_ITEM(args, i, arg);
+	}
+	if (!tuple_success) {
+		Py_DECREF(args);
+		end_call_python(runtime_data);
+		return JS_ThrowInternalError(ctx, "Internal error: could not convert args.");
+	}
+
+	PyObject *result = PyObject_CallObject(node->obj, args);
+	Py_DECREF(args);
+	if (!result) {
+		end_call_python(runtime_data);
+		return JS_ThrowInternalError(ctx, "Python call failed.");
+	}
+	JSValue js_result = JS_NULL;
+	if (python_to_quickjs_possible(runtime_data, result)) {
+		js_result = python_to_quickjs(runtime_data, result);
+	} else {
+		PyErr_Clear();
+		js_result = JS_ThrowInternalError(ctx, "Can not convert Python result to JS.");
+	}
+	Py_DECREF(result);
+
+	end_call_python(runtime_data);
+	return js_result;
+}
+
+static JSClassDef js_python_function_class = {
+	"PythonFunction",
+	.finalizer = js_python_function_finalizer,
+	.call = js_python_function_call,
+};
+
 // Creates an instance of the _quickjs.Context class.
-static PyObject *context_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
-	ContextData *self = PyObject_GC_New(ContextData, type);
+static PyObject *runtime_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
+	RuntimeData *self = PyObject_GC_New(RuntimeData, type);
 	if (self != NULL) {
 		// We never have different contexts for the same runtime. This way, different
 		// _quickjs.Context can be used concurrently.
 		self->runtime = JS_NewRuntime();
 		self->context = JS_NewContext(self->runtime);
+		JS_NewClass(self->runtime, js_python_function_class_id,
+		            &js_python_function_class);
+		JSValue global = JS_GetGlobalObject(self->context);
+		JSValue fct_cls = JS_GetPropertyStr(self->context, global, "Function");
+		JSValue fct_proto = JS_GetPropertyStr(self->context, fct_cls, "prototype");
+		JS_FreeValue(self->context, fct_cls);
+		JS_SetClassProto(self->context, js_python_function_class_id, fct_proto);
+		JS_FreeValue(self->context, global);
 		self->has_time_limit = 0;
 		self->time_limit = 0;
 		self->thread_state = NULL;
 		self->python_callables = NULL;
-		JS_SetContextOpaque(self->context, self);
+		JS_SetRuntimeOpaque(self->runtime, self);
 		PyObject_GC_Track(self);
 	}
 	return (PyObject *)self;
 }
 
 // Deallocates an instance of the _quickjs.Context class.
-static void context_dealloc(ContextData *self) {
+static void runtime_dealloc(RuntimeData *self) {
 	JS_FreeContext(self->context);
 	JS_FreeRuntime(self->runtime);
 	PyObject_GC_UnTrack(self);
-	PythonCallableNode *node = self->python_callables;
-	self->python_callables = NULL;
-	while (node) {
-		PythonCallableNode *this = node;
-		node = node->next;
-		// this->obj may already be NULL if GC'ed right before through context_clear.
-		Py_XDECREF(this->obj);
-		PyMem_Free(this);
-	}
 	PyObject_GC_Del(self);
 }
 
 // Evaluates a Python string as JS and returns the result as a Python object. Will return
 // _quickjs.Object for complex types (other than e.g. str, int).
-static PyObject *context_eval_internal(ContextData *self, PyObject *args, int eval_type) {
+static PyObject *runtime_eval_internal(RuntimeData *self, PyObject *args, int eval_type) {
 	const char *code;
 	if (!PyArg_ParseTuple(args, "s", &code)) {
 		return NULL;
@@ -416,21 +514,21 @@ static PyObject *context_eval_internal(ContextData *self, PyObject *args, int ev
 //
 // Evaluates a Python string as JS and returns the result as a Python object. Will return
 // _quickjs.Object for complex types (other than e.g. str, int).
-static PyObject *context_eval(ContextData *self, PyObject *args) {
-	return context_eval_internal(self, args, JS_EVAL_TYPE_GLOBAL);
+static PyObject *runtime_eval(RuntimeData *self, PyObject *args) {
+	return runtime_eval_internal(self, args, JS_EVAL_TYPE_GLOBAL);
 }
 
 // _quickjs.Context.module
 //
 // Evaluates a Python string as JS module. Otherwise identical to eval.
-static PyObject *context_module(ContextData *self, PyObject *args) {
-	return context_eval_internal(self, args, JS_EVAL_TYPE_MODULE);
+static PyObject *runtime_module(RuntimeData *self, PyObject *args) {
+	return runtime_eval_internal(self, args, JS_EVAL_TYPE_MODULE);
 }
 
 // _quickjs.Context.execute_pending_job
 //
 // If there are pending jobs, executes one and returns True. Else returns False.
-static PyObject *context_execute_pending_job(ContextData *self) {
+static PyObject *runtime_execute_pending_job(RuntimeData *self) {
 	prepare_call_js(self);
 	JSContext *ctx;
 	int ret = JS_ExecutePendingJob(self->runtime, &ctx);
@@ -449,14 +547,14 @@ static PyObject *context_execute_pending_job(ContextData *self) {
 //
 // Evaluates a Python string as JSON and returns the result as a Python object. Will
 // return _quickjs.Object for complex types (other than e.g. str, int).
-static PyObject *context_parse_json(ContextData *self, PyObject *args) {
+static PyObject *runtime_parse_json(RuntimeData *self, PyObject *args) {
 	const char *data;
 	if (!PyArg_ParseTuple(args, "s", &data)) {
 		return NULL;
 	}
 	JSValue value;
 	Py_BEGIN_ALLOW_THREADS;
-	value = JS_ParseJSON(self->context, data, strlen(data), "context_parse_json.json");
+	value = JS_ParseJSON(self->context, data, strlen(data), "runtime_parse_json.json");
 	Py_END_ALLOW_THREADS;
 	return quickjs_to_python(self, value);
 }
@@ -464,7 +562,7 @@ static PyObject *context_parse_json(ContextData *self, PyObject *args) {
 // _quickjs.Context.get
 //
 // Retrieves a global variable from the JS context.
-static PyObject *context_get(ContextData *self, PyObject *args) {
+static PyObject *runtime_get(RuntimeData *self, PyObject *args) {
 	const char *name;
 	if (!PyArg_ParseTuple(args, "s", &name)) {
 		return NULL;
@@ -478,7 +576,7 @@ static PyObject *context_get(ContextData *self, PyObject *args) {
 // _quickjs.Context.set
 //
 // Sets a global variable to the JS context.
-static PyObject *context_set(ContextData *self, PyObject *args) {
+static PyObject *runtime_set(RuntimeData *self, PyObject *args) {
 	const char *name;
 	PyObject *item;
 	if (!PyArg_ParseTuple(args, "sO", &name, &item)) {
@@ -503,7 +601,7 @@ static PyObject *context_set(ContextData *self, PyObject *args) {
 // _quickjs.Context.set_memory_limit
 //
 // Sets the memory limit of the context.
-static PyObject *context_set_memory_limit(ContextData *self, PyObject *args) {
+static PyObject *runtime_set_memory_limit(RuntimeData *self, PyObject *args) {
 	Py_ssize_t limit;
 	if (!PyArg_ParseTuple(args, "n", &limit)) {
 		return NULL;
@@ -515,7 +613,7 @@ static PyObject *context_set_memory_limit(ContextData *self, PyObject *args) {
 // _quickjs.Context.set_time_limit
 //
 // Sets the CPU time limit of the context. This will be used in an interrupt handler.
-static PyObject *context_set_time_limit(ContextData *self, PyObject *args) {
+static PyObject *runtime_set_time_limit(RuntimeData *self, PyObject *args) {
 	double limit;
 	if (!PyArg_ParseTuple(args, "d", &limit)) {
 		return NULL;
@@ -532,7 +630,7 @@ static PyObject *context_set_time_limit(ContextData *self, PyObject *args) {
 // _quickjs.Context.set_max_stack_size
 //
 // Sets the max stack size in bytes.
-static PyObject *context_set_max_stack_size(ContextData *self, PyObject *args) {
+static PyObject *runtime_set_max_stack_size(RuntimeData *self, PyObject *args) {
 	Py_ssize_t limit;
 	if (!PyArg_ParseTuple(args, "n", &limit)) {
 		return NULL;
@@ -544,7 +642,7 @@ static PyObject *context_set_max_stack_size(ContextData *self, PyObject *args) {
 // _quickjs.Context.memory
 //
 // Sets the CPU time limit of the context. This will be used in an interrupt handler.
-static PyObject *context_memory(ContextData *self) {
+static PyObject *runtime_memory(RuntimeData *self) {
 	PyObject *dict = PyDict_New();
 	if (dict == NULL) {
 		return NULL;
@@ -591,66 +689,13 @@ static PyObject *context_memory(ContextData *self) {
 // _quickjs.Context.gc
 //
 // Runs garbage collection.
-static PyObject *context_gc(ContextData *self) {
+static PyObject *runtime_gc(RuntimeData *self) {
 	JS_RunGC(self->runtime);
 	Py_RETURN_NONE;
 }
 
-static JSValue js_c_function(
-    JSContext *ctx, JSValueConst this_val, int argc, JSValueConst *argv, int magic) {
-	ContextData *context = (ContextData *)JS_GetContextOpaque(ctx);
-	if (context->has_time_limit) {
-		return JS_ThrowInternalError(ctx, "Can not call into Python with a time limit set.");
-	}
-	PythonCallableNode *node = context->python_callables;
-	while (node && node->magic != magic) {
-		node = node->next;
-	}
-	if (!node) {
-		return JS_ThrowInternalError(ctx, "Internal error.");
-	}
-	prepare_call_python(context);
 
-	PyObject *args = PyTuple_New(argc);
-	if (!args) {
-		end_call_python(context);
-		return JS_ThrowOutOfMemory(ctx);
-	}
-	int tuple_success = 1;
-	for (int i = 0; i < argc; ++i) {
-		PyObject *arg = quickjs_to_python(context, JS_DupValue(ctx, argv[i]));
-		if (!arg) {
-			tuple_success = 0;
-			break;
-		}
-		PyTuple_SET_ITEM(args, i, arg);
-	}
-	if (!tuple_success) {
-		Py_DECREF(args);
-		end_call_python(context);
-		return JS_ThrowInternalError(ctx, "Internal error: could not convert args.");
-	}
-
-	PyObject *result = PyObject_CallObject(node->obj, args);
-	Py_DECREF(args);
-	if (!result) {
-		end_call_python(context);
-		return JS_ThrowInternalError(ctx, "Python call failed.");
-	}
-	JSValue js_result = JS_NULL;
-	if (python_to_quickjs_possible(context, result)) {
-		js_result = python_to_quickjs(context, result);
-	} else {
-		PyErr_Clear();
-		js_result = JS_ThrowInternalError(ctx, "Can not convert Python result to JS.");
-	}
-	Py_DECREF(result);
-
-	end_call_python(context);
-	return js_result;
-}
-
-static PyObject *context_add_callable(ContextData *self, PyObject *args) {
+static PyObject *runtime_add_callable(RuntimeData *self, PyObject *args) {
 	const char *name;
 	PyObject *callable;
 	if (!PyArg_ParseTuple(args, "sO", &name, &callable)) {
@@ -661,27 +706,34 @@ static PyObject *context_add_callable(ContextData *self, PyObject *args) {
 		return NULL;
 	}
 
+	JSValue function = JS_NewObjectClass(self->context, js_python_function_class_id);
+	if (JS_IsException(function)) {
+		quickjs_exception_to_python(self->context);
+		return NULL;
+	}
+	// TODO: Should we allow setting the .length of the function to something other than 0?
+	JS_DefinePropertyValueStr(self->context, function, "name", JS_NewString(self->context, name), JS_PROP_CONFIGURABLE);
 	PythonCallableNode *node = PyMem_Malloc(sizeof(PythonCallableNode));
 	if (!node) {
+		JS_FreeValue(self->context, function);
 		return NULL;
 	}
 	Py_INCREF(callable);
-	node->magic = 0;
-	if (self->python_callables) {
-		node->magic = self->python_callables->magic + 1;
-	}
 	node->obj = callable;
+	node->prev = NULL;
 	node->next = self->python_callables;
+	if (self->python_callables) {
+		self->python_callables->prev = node;
+	}
 	self->python_callables = node;
+	JS_SetOpaque(function, node);
 
-	JSValue function = JS_NewCFunctionMagic(
-	    self->context,
-	    js_c_function,
-	    name,
-	    0,  // TODO: Should we allow setting the .length of the function to something other than 0?
-	    JS_CFUNC_generic_magic,
-	    node->magic);
 	JSValue global = JS_GetGlobalObject(self->context);
+	if (JS_IsException(global)) {
+		JS_FreeValue(self->context, function);
+		quickjs_exception_to_python(self->context);
+		return NULL;
+	}
 	// If this fails we don't notify the caller of this function.
 	int ret = JS_SetPropertyStr(self->context, global, name, function);
 	JS_FreeValue(self->context, global);
@@ -693,46 +745,62 @@ static PyObject *context_add_callable(ContextData *self, PyObject *args) {
 	}
 }
 
+
+// _quickjs.Context.globalThis
+//
+// Global object of the JS context.
+static PyObject *runtime_global_this(RuntimeData *self, void *closure) {
+	return quickjs_to_python(self, JS_GetGlobalObject(self->context));
+}
+
+
 // All methods of the _quickjs.Context class.
-static PyMethodDef context_methods[] = {
-    {"eval", (PyCFunction)context_eval, METH_VARARGS, "Evaluates a Javascript string."},
+static PyMethodDef runtime_methods[] = {
+    {"eval", (PyCFunction)runtime_eval, METH_VARARGS, "Evaluates a Javascript string."},
     {"module",
-     (PyCFunction)context_module,
+     (PyCFunction)runtime_module,
      METH_VARARGS,
      "Evaluates a Javascript string as a module."},
-    {"execute_pending_job", (PyCFunction)context_execute_pending_job, METH_NOARGS, "Executes a pending job."},
-    {"parse_json", (PyCFunction)context_parse_json, METH_VARARGS, "Parses a JSON string."},
-    {"get", (PyCFunction)context_get, METH_VARARGS, "Gets a Javascript global variable."},
-    {"set", (PyCFunction)context_set, METH_VARARGS, "Sets a Javascript global variable."},
+    {"execute_pending_job", (PyCFunction)runtime_execute_pending_job, METH_NOARGS, "Executes a pending job."},
+    {"parse_json", (PyCFunction)runtime_parse_json, METH_VARARGS, "Parses a JSON string."},
+    {"get", (PyCFunction)runtime_get, METH_VARARGS, "Gets a Javascript global variable."},
+    {"set", (PyCFunction)runtime_set, METH_VARARGS, "Sets a Javascript global variable."},
     {"set_memory_limit",
-     (PyCFunction)context_set_memory_limit,
+     (PyCFunction)runtime_set_memory_limit,
      METH_VARARGS,
      "Sets the memory limit in bytes."},
     {"set_time_limit",
-     (PyCFunction)context_set_time_limit,
+     (PyCFunction)runtime_set_time_limit,
      METH_VARARGS,
      "Sets the CPU time limit in seconds (C function clock() is used)."},
     {"set_max_stack_size",
-     (PyCFunction)context_set_max_stack_size,
+     (PyCFunction)runtime_set_max_stack_size,
      METH_VARARGS,
      "Sets the maximum stack size in bytes. Default is 256kB."},
-    {"memory", (PyCFunction)context_memory, METH_NOARGS, "Returns the memory usage as a dict."},
-    {"gc", (PyCFunction)context_gc, METH_NOARGS, "Runs garbage collection."},
-    {"add_callable", (PyCFunction)context_add_callable, METH_VARARGS, "Wraps a Python callable."},
+    {"memory", (PyCFunction)runtime_memory, METH_NOARGS, "Returns the memory usage as a dict."},
+    {"gc", (PyCFunction)runtime_gc, METH_NOARGS, "Runs garbage collection."},
+    {"add_callable", (PyCFunction)runtime_add_callable, METH_VARARGS, "Wraps a Python callable."},
+    {NULL} /* Sentinel */
+};
+
+// All getsetters (properties) of the _quickjs.Context class.
+static PyGetSetDef runtime_getsetters[] = {
+    {"globalThis", (getter)runtime_global_this, NULL, "Global object of the context.", NULL},
     {NULL} /* Sentinel */
 };
 
 // Define the _quickjs.Context type.
 static PyTypeObject Context = {PyVarObject_HEAD_INIT(NULL, 0).tp_name = "_quickjs.Context",
                                .tp_doc = "Quickjs context",
-                               .tp_basicsize = sizeof(ContextData),
+                               .tp_basicsize = sizeof(RuntimeData),
                                .tp_itemsize = 0,
                                .tp_flags = Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
-                               .tp_traverse = (traverseproc)context_traverse,
-                               .tp_clear = (inquiry)context_clear,
-                               .tp_new = context_new,
-                               .tp_dealloc = (destructor)context_dealloc,
-                               .tp_methods = context_methods};
+                               .tp_traverse = (traverseproc)runtime_traverse,
+                               .tp_clear = (inquiry)runtime_clear,
+                               .tp_new = runtime_new,
+                               .tp_dealloc = (destructor)runtime_dealloc,
+                               .tp_methods = runtime_methods,
+                               .tp_getset = runtime_getsetters};
 
 // All global methods in _quickjs.
 static PyMethodDef myextension_methods[] = {{"test", (PyCFunction)test, METH_NOARGS, NULL},
@@ -763,6 +831,8 @@ PyMODINIT_FUNC PyInit__quickjs(void) {
 		return NULL;
 	}
 
+	JS_NewClassID(&js_python_function_class_id);
+
 	JSException = PyErr_NewException("_quickjs.JSException", NULL, NULL);
 	if (JSException == NULL) {
 		return NULL;
